"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var path_1 = __importDefault(require("path"));
var typescript_1 = __importDefault(require("typescript"));
var url_1 = __importDefault(require("url"));
// endregion
/* ****************************************************************************************************************** */
// region: Helpers
/* ****************************************************************************************************************** */
var getImplicitExtensions = function (options) {
    var res = [".ts", ".d.ts"];
    var allowJs = options.allowJs, jsx = options.jsx, allowJson = options.resolveJsonModule;
    var allowJsx = !!jsx && jsx !== typescript_1.default.JsxEmit.None;
    allowJs && res.push(".js");
    allowJsx && res.push(".tsx");
    allowJs && allowJsx && res.push(".jsx");
    allowJson && res.push(".json");
    return res;
};
var isURL = function (s) {
    return !!s && (!!url_1.default.parse(s).host || !!url_1.default.parse(s).hostname);
};
var isBaseDir = function (base, dir) { var _a; return ((_a = path_1.default.relative(base, dir)) === null || _a === void 0 ? void 0 : _a[0]) !== "."; };
var isRequire = function (node) {
    return typescript_1.default.isCallExpression(node) &&
        typescript_1.default.isIdentifier(node.expression) &&
        node.expression.text === "require" &&
        typescript_1.default.isStringLiteral(node.arguments[0]) &&
        node.arguments.length === 1;
};
var isAsyncImport = function (node) {
    return typescript_1.default.isCallExpression(node) &&
        node.expression.kind === typescript_1.default.SyntaxKind.ImportKeyword &&
        typescript_1.default.isStringLiteral(node.arguments[0]) &&
        node.arguments.length === 1;
};
// endregion
/* ****************************************************************************************************************** *
 * Transformer
 * ****************************************************************************************************************** */
function transformer(program, config) {
    var useRootDirs = config.useRootDirs;
    var compilerOptions = program.getCompilerOptions();
    var implicitExtensions = getImplicitExtensions(compilerOptions);
    return function (context) { return function (sourceFile) {
        var _a;
        // TS 4 - new node factory
        var factory = context.factory;
        var fileName = sourceFile.fileName;
        var fileDir = typescript_1.default.normalizePath(path_1.default.dirname(fileName));
        var baseUrl = compilerOptions.baseUrl;
        if (!baseUrl || !compilerOptions.paths)
            return sourceFile;
        var rootDirs = (_a = compilerOptions.rootDirs) === null || _a === void 0 ? void 0 : _a.filter(path_1.default.isAbsolute);
        return typescript_1.default.visitEachChild(sourceFile, visit, context);
        /* ********************************************************* *
         * Transformer Helpers
         * ********************************************************* */
        /**
         * Gets proper path and calls updaterFn to update the node
         */
        function update(original, moduleName, updaterFn) {
            var p;
            /* Have Compiler API attempt to resolve */
            var _a = typescript_1.default.resolveModuleName(moduleName, fileName, compilerOptions, typescript_1.default.sys), resolvedModule = _a.resolvedModule, failedLookupLocations = _a.failedLookupLocations;
            if (!resolvedModule) {
                var maybeURL = failedLookupLocations[0];
                if (!isURL(maybeURL))
                    return original;
                p = maybeURL;
            }
            else if (resolvedModule.isExternalLibraryImport)
                return original;
            else {
                var extension = resolvedModule.extension, resolvedFileName = resolvedModule.resolvedFileName;
                var filePath = fileDir;
                var modulePath = path_1.default.dirname(resolvedFileName);
                /* Handle rootDirs mapping */
                if (useRootDirs && rootDirs) {
                    var fileRootDir = "";
                    var moduleRootDir = "";
                    for (var _i = 0, rootDirs_1 = rootDirs; _i < rootDirs_1.length; _i++) {
                        var rootDir = rootDirs_1[_i];
                        if (isBaseDir(rootDir, resolvedFileName) &&
                            rootDir.length > moduleRootDir.length)
                            moduleRootDir = rootDir;
                        if (isBaseDir(rootDir, fileName) &&
                            rootDir.length > fileRootDir.length)
                            fileRootDir = rootDir;
                    }
                    /* Remove base dirs to make relative to root */
                    if (fileRootDir && moduleRootDir) {
                        filePath = path_1.default.relative(fileRootDir, filePath);
                        modulePath = path_1.default.relative(moduleRootDir, modulePath);
                    }
                }
                /* Remove extension if implicit */
                p = typescript_1.default.normalizePath(path_1.default.join(path_1.default.relative(filePath, modulePath), path_1.default.basename(resolvedFileName)));
                if (extension && implicitExtensions.includes(extension))
                    p = p.slice(0, -extension.length);
                if (!p)
                    return original;
                p = p[0] === "." ? p : "./" + p;
            }
            return updaterFn(typescript_1.default.createLiteral(p));
        }
        /**
         * Visit and replace nodes with module specifiers
         */
        function visit(node) {
            /* Update require() or import() */
            if (isRequire(node) || isAsyncImport(node))
                return update(node, node.arguments[0].text, function (p) {
                    return factory
                        ? factory.updateCallExpression(node, node.expression, node.typeArguments, [p])
                        : typescript_1.default.updateCall(node, node.expression, node.typeArguments, [p]);
                });
            /* Update ExternalModuleReference - import foo = require("foo"); */
            if (typescript_1.default.isExternalModuleReference(node) &&
                typescript_1.default.isStringLiteral(node.expression))
                return update(node, node.expression.text, function (p) {
                    return factory
                        ? factory.updateExternalModuleReference(node, p)
                        : typescript_1.default.updateExternalModuleReference(node, p);
                });
            /**
             * Update ImportDeclaration / ExportDeclaration
             * import ... 'module';
             * export ... 'module';
             *
             * This implements a workaround for the following TS issues:
             * @see https://github.com/microsoft/TypeScript/issues/40603
             * @see https://github.com/microsoft/TypeScript/issues/31446
             */
            if ((typescript_1.default.isImportDeclaration(node) || typescript_1.default.isExportDeclaration(node)) &&
                node.moduleSpecifier &&
                typescript_1.default.isStringLiteral(node.moduleSpecifier))
                return update(node, node.moduleSpecifier.text, function (p) {
                    return factory
                        ? Object.assign(node, {
                            moduleSpecifier: p,
                        })
                        : Object.assign(node, {
                            moduleSpecifier: typescript_1.default.updateNode(p, node.moduleSpecifier),
                        });
                });
            /* Update ImportTypeNode - typeof import("./bar"); */
            if (typescript_1.default.isImportTypeNode(node)) {
                var argument_1 = node.argument;
                if (!typescript_1.default.isStringLiteral(argument_1.literal))
                    return node;
                var text = argument_1.literal.text;
                return !text
                    ? node
                    : update(node, text, function (p) {
                        return factory
                            ? factory.updateImportTypeNode(node, factory.updateLiteralTypeNode(argument_1, p), node.qualifier, node.typeArguments, node.isTypeOf)
                            : typescript_1.default.updateImportTypeNode(node, typescript_1.default.updateLiteralTypeNode(argument_1, p), node.qualifier, node.typeArguments, node.isTypeOf);
                    });
            }
            return typescript_1.default.visitEachChild(node, visit, context);
        }
    }; };
}
exports.default = transformer;
